from typing import Dict
from fastapi import FastAPI, HTTPException, Response
from uuid import UUID, uuid4
from pydantic import BaseModel
from starlette import status
from starlette.responses import FileResponse, RedirectResponse
from models import RevenueMaximisationJob, JobStatus, Job, Solution, MarkdownOptimisationJob, JobType
from concurrent.futures import ThreadPoolExecutor, Future

import logging
import numpy as np
import revenue_maximization.revenue as revenue_mip
import markdown_optimisation.markdown as markdown_mip

description = """
Price Optimisation Engine API can be used to solve awesome price optimisation problems. 🚀

## Perishable product with fixed capacity and price flexibility

This problem setting is relevant for **air tickets, hotel rooms, car rentals, theater tickets**. 
* Maximize revenue generated (from a single product)
* Single product (e.g. hotel rooms of a calendar date, tickets of a soccer game)
* Finite horizon (lead time) until the product is consumed (e.g. the date of the flight).
* Fixed capacity (initial inventory) at period zero is given
* Pricing is flexible
* Product has zero value after consumption date

_Input data:_
* Lead time (this will translate to a planning horizon T)
* Price ladder (discrete set of prices P). Prices must be given in increasing order.
* Demand matrix (forecasted demand per price per period – preprocessed). Rows correspond to prices, columns to periods.

_Decision variables:_
* Binary variable x_{tp} (1 if price p \in P is selected for period t \in T; otherwise 0)
* Continuous variable y_{tp} number of items sold at period t \in T


## Markdown optimization for perishable product with fixed capacity and salvage value (Version 1)

This problem setting is relevant for **apparel, style and seasonal-goods, high-tech and perishable food**. This version 
assumes that the demand matrix is preprocessed and provided as an input (demand for each price point per period). From
the size of the demand matrix the planning horizon is inferred. 

* Maximize revenue of the sales season (from a single product). 
* Product perishes and has a low salvage value once the season is over. 
* Pricing is not flexible and only involves price reductions. 
* Markdown is needed to manage the demand and maximize revenue 
* Every period the retailer re-runs the price optimization engine with updated info. 

_Input data:_
* Salvage price s
* Price ladder (discrete set of prices P). Values must be given with decreasing order.
* Initial inventory I_{0}
* Demand matrix (forecasted demand per price per period – preprocessed). Rows correspond to prices, columns to periods.

_Decision variables:_
* Binary variable x_{tp} (1 if price p is selected for period t; otherwise 0)
* Continuous variables I_{T} refers to the inventory left at the end of the season


The CBC_COIN_OR open-source solver is used for solving the mathematical programming models.
"""

app = FastAPI(
    title="Satalia's Price Optimisation Engine",
    description=description,
    version="0.0.1",
    # TODO: select terms of usage for API, if applicable
    contact={
        "name": "Satalia",
        "url": "https://www.satalia.com/contact/",
        "email": "hello@satalia.com",
    },
    # TODO: select correct license for API, if applicable
)

dbJobs: Dict[UUID, Job] = {}
dbFutures: Dict[UUID, Future] = {}
executor = ThreadPoolExecutor(max_workers=3)
logger = logging.getLogger(__name__)


class Message(BaseModel):
    message: str


@app.get('/favicon.ico', include_in_schema=False)
async def favicon():
    return FileResponse('favicon.ico')

@app.get("/", include_in_schema=False)
async def docs_redirect():
    return RedirectResponse(url='/docs')

@app.get('/_ah/warmup', include_in_schema=False, status_code=status.HTTP_200_OK)
def warmup():
    return {}

@app.get("/v1/health", status_code=status.HTTP_200_OK)
async def root():
    return {"message": "Satalia's Price Optimisation Engines Health Check"}


@app.get("/v1/jobs/", status_code=status.HTTP_200_OK)
async def return_jobs():
    return dbJobs


@app.get("/v1/jobs/solution/{uuid}",
         response_model=Solution,
         responses={404: {"model": Message}, 425: {"model": Message}})
async def return_solution(uuid: UUID):
    for jobID, future in dbFutures.items():
        if jobID == uuid:
            if future.done():
                return future.result()
            if future.cancelled() or future.running():
                raise HTTPException(
                    status_code=425,
                    detail=f"Solution future is empty, job {uuid} is queued or running"
                )
    raise HTTPException(
        status_code=404,
        detail=f"no solution exist for job id: {uuid}"
    )


@app.delete("/v1/jobs/solution/{uuid}",
            status_code=status.HTTP_204_NO_CONTENT,
            responses={404: {"model": Message}},
            response_class=Response)
async def delete_solution(uuid: UUID):
    for jobID, future in dbFutures.items():
        if jobID == uuid:
            future.cancel()
            del dbJobs[uuid]
            del dbFutures[uuid]
            return
    raise HTTPException(
        status_code=404,
        detail=f"no solution exist for job id: {uuid}"
    )


@app.get("/v1/jobs/status/{uuid}", responses={404: {"model": Message}})
async def return_job_status(uuid: UUID):
    for jobID, future in dbFutures.items():
        if jobID == uuid:
            if future.done():
                return {"status": JobStatus.finished}
            if future.cancelled():
                return {"status": JobStatus.cancelled}
            if future.running():
                return {"status": JobStatus.running}
    raise HTTPException(
        status_code=404,
        detail=f"job with id: {uuid} does not exist"
    )


@app.post("/v1/revenue", responses={400: {"model": Message}})
async def submit_revenue_job(job: RevenueMaximisationJob):
    job.id = uuid4()
    job.type = JobType.revenue
    dbJobs[job.id] = job
    demand_matrix = np.array(job.demand_matrix)

    try:
        revenue_mip.input_validation(job.capacity, demand_matrix, job.prices)
    except ValueError as er:
        logger.info(f'Input data error: {er}')
        raise HTTPException(
            status_code=400,
            detail=f"Input data error: {er}")

    dbFutures[job.id] = executor.submit(
        revenue_mip.optimal_prices, job.capacity, demand_matrix, job.prices)
    return {"id": job.id}


@app.post("/v1/markdown", responses={400: {"model": Message}})
async def submit_markdown_job(job: MarkdownOptimisationJob):
    job.id = uuid4()
    job.type = JobType.markdown
    dbJobs[job.id] = job
    demand_matrix = np.array(job.demand_matrix)

    try:
        markdown_mip.input_validation(job.initial_inventory, demand_matrix, job.prices, job.salvage_value)
    except ValueError as er:
        logger.info(f'Input data error: {er}')
        raise HTTPException(
            status_code=400,
            detail=f"Input data error: {er}")

    dbFutures[job.id] = executor.submit(
        markdown_mip.optimal_prices, job.initial_inventory, demand_matrix, job.prices, job.salvage_value)
    return {"id": job.id}
