import logging
from typing import List
from models import Solution

import numpy as np
import pulp


def input_validation(
        initial_inventory: int,
        demand_matrix: np.ndarray,
        prices: List[float],
        salvage_value: float
):
    """
        This function checks the input data

        :param salvage_value: salvage value of the product at the end of the season
        :param initial_inventory: number of available items at the start of the season
        :param demand_matrix: 2d demand array of price (rows) and periods (columns)
        :param prices: list of prices to select from
    """

    if demand_matrix.size == 0:
        raise ValueError('demand matrix must be provided')

    if initial_inventory < 1:
        raise ValueError('initial inventory should be greater than 1')

    if salvage_value < 0:
        raise ValueError('salvage value should be greater than 0')

    n_prices, n_periods = demand_matrix.shape

    if n_prices != len(prices):
        raise ValueError('Demand matrix does not match prices length')

    if (np.diff(demand_matrix, axis=0) < 0).any():
        raise ValueError('Demand matrix is inconsistent (price-demand have positive correlation.')


def optimal_prices(
        initial_inventory: int,
        demand_matrix: np.ndarray,
        prices: List[float],
        salvage_value: float
) -> Solution:
    """
    This function models and solves the single product markdown optimisation problem.

        :param salvage_value: salvage value of the product at the end of the season
        :param initial_inventory: number of available items at the start of the season
        :param demand_matrix: 2d demand array of price (rows) and periods (columns)
        :param prices: list of prices to select from
    """

    logger = logging.getLogger(__name__)
    n_prices, n_periods = demand_matrix.shape
    indices = [
        (p, t)
        for p in range(n_prices)
        for t in range(n_periods)
    ]

    xi = 0.999
    max_demands = demand_matrix.max(axis=0).clip(-np.inf, initial_inventory)
    big_m1 = max_demands[:-1]
    big_m2 = max_demands[1:]

    # Variables
    x_var = pulp.LpVariable.dicts(
        'x',
        indices,
        cat=pulp.LpBinary
    )

    z_var = pulp.LpVariable.dicts(
        'z',
        indices,
        lowBound=0,
        cat=pulp.LpContinuous
    )

    i_var = pulp.LpVariable('i', lowBound=0, cat='Continuous')

    a_var = pulp.LpVariable.dicts(
        'a',
        range(n_periods),
        cat=pulp.LpBinary
    )

    # direction
    model = pulp.LpProblem('Markdown_Optimisation_MILP', pulp.LpMaximize)

    # objective function --> revenue
    model += pulp.lpSum(
        z_var[p, t] * prices[p]
        for p, t in indices
    ) + i_var * salvage_value

    # One price should be selected for every period
    for t in range(n_periods):
        model += pulp.lpSum(
            x_var[p, t]
            for p in range(n_prices)
        ) == 1

    # Products sold and remaining inventory should be equal to the initial inventory
    model += pulp.lpSum(z_var[i] for i in indices) + i_var == initial_inventory

    # definition of z variable (z=x*y)
    for i in indices:
        model += z_var[i] <= x_var[i] * demand_matrix[i]

    # Start at period 0 with the highest price
    model += x_var[0, 0] == 1

    # ensure price decreasing consistency
    for p in range(n_prices):
        for t in range(n_periods - 1):
            model += \
                pulp.lpSum(x_var[q, t] for q in range(p + 1)) - pulp.lpSum(x_var[q, t + 1] for q in range(p + 1)) >= 0

    # if items are fully consumed before the end of the planning horizon, z must be forced to zero
    for t in range(n_periods - 1):
        model += -big_m1[t] * a_var[t] <= pulp.lpSum(
            z_var[p, t] - xi * demand_matrix[(p, t)] * x_var[(p, t)]
            for p in range(n_prices)
        )
        model += a_var[t] >= pulp.lpSum(
            z_var[p, t] - demand_matrix[(p, t)] * x_var[(p, t)]
            for p in range(n_prices)
        )
        model += pulp.lpSum(
            z_var[p, t + 1]
            for p in range(n_prices)
        ) <= big_m2[t] * (1 - a_var[t])

    # solve model
    model.solve(pulp.PULP_CBC_CMD(fracGap=0.01))

    if model.status == 1:
        logger.info(f'Optimal Solution Found: {pulp.LpStatus[model.status]}')
        logger.info(f'Objective Function: {pulp.value(model.objective)}')
    else:
        logger.debug(f'Price Optimization Failed - status: {pulp.LpStatus[model.status]}')

    eps = 1.e-6  # used to avoid numerical weirdness in the solver's output
    best_price = []
    for t in range(n_periods):
        best_price.append(
            next((prices[p]
                  for p in range(n_prices)
                  if x_var[p, t].varValue > eps and z_var[p, t].varValue > eps), prices[0]))

    return Solution(optimal_prices=best_price, objective=pulp.value(model.objective), status=model.status)
