import logging
from typing import List
from models import Solution

import numpy as np
import pulp


def input_validation(
        capacity: int,
        demand_matrix: np.ndarray,
        prices: List[float],
):
    """
        This function checks the input data

            :param capacity: number of available rooms for given booking night
            :param demand_matrix: 2d demand array of price (rows) and time horizon (columns)
            :param prices: list of prices to select from
    """
    if demand_matrix.size == 0:
        raise ValueError('needs predictions for calculation')

    if capacity < 1:
        raise ValueError('available capacity should be greater than 0')

    n_prices, n_periods = demand_matrix.shape

    if n_prices != len(prices):
        raise ValueError('Demand matrix does not match prices length')

    if (np.diff(demand_matrix, axis=0) > 0).any():
        raise ValueError('Demand matrix is inconsistent (price-demand have positive correlation.')


def optimal_prices(
        capacity: int,
        demand_matrix: np.ndarray,
        prices: List[float],
) -> Solution:
    """
    This function uses a mixed integer linear programming approach to find the best prices that maximize revenue.

        :param capacity: number of available rooms for given booking night
        :param demand_matrix: 2d demand array of price (rows) and time horizon (columns)
        :param prices: list of prices to select from
    """

    logger = logging.getLogger(__name__)

    n_prices, n_periods = demand_matrix.shape

    indices = [
        (p, t)
        for p in range(n_prices)
        for t in range(n_periods)
    ]

    xi = 0.999
    max_demands = demand_matrix.max(axis=0).clip(-np.inf, capacity)
    big_m1 = max_demands[:-1]
    big_m2 = max_demands[1:]

    # Variables
    x_var = pulp.LpVariable.dicts(
        'x',
        indices,
        cat=pulp.LpBinary
    )

    z_var = pulp.LpVariable.dicts(
        'z',
        indices,
        lowBound=0,
        cat=pulp.LpContinuous
    )

    a_var = pulp.LpVariable.dicts(
        'a',
        range(n_periods),
        cat=pulp.LpBinary
    )

    # direction
    model = pulp.LpProblem('Revenue_MILP', pulp.LpMaximize)

    # objective function --> revenue
    model += pulp.lpSum(
        z_var[p, lead_time] * prices[p]
        for p, lead_time in indices
    )

    # One price should be selected for every period
    for lead_time in range(n_periods):
        model += pulp.lpSum(
            x_var[p, lead_time]
            for p in range(n_prices)
        ) == 1

    # Rooms sold must not exceed given capacity
    model += pulp.lpSum(z_var[i] for i in indices) <= capacity

    # definition of z variable (z=x*y)
    for i in indices:
        model += z_var[i] <= x_var[i] * demand_matrix[i]

    # if rooms are fully consumed before the end of the planning horizon, z must be forced to zero
    for lead_time in range(n_periods - 1):
        model += -big_m1[lead_time] * a_var[lead_time] <= pulp.lpSum(
            z_var[p, lead_time] - xi * demand_matrix[(p, lead_time)] * x_var[(p, lead_time)]
            for p in range(n_prices)
        )
        model += a_var[lead_time] >= pulp.lpSum(
            z_var[p, lead_time] - demand_matrix[(p, lead_time)] * x_var[(p, lead_time)]
            for p in range(n_prices)
        )
        model += pulp.lpSum(
            z_var[p, lead_time + 1]
            for p in range(n_prices)
        ) <= big_m2[lead_time] * (1 - a_var[lead_time])

    # solve model
    model.solve(pulp.PULP_CBC_CMD(fracGap=0.01))

    if model.status == 1:
        logger.info(f'Optimal Solution Found: {pulp.LpStatus[model.status]}')
        logger.info(f'Objective Function: {pulp.value(model.objective)}')
    else:
        logger.debug(f'Optimization Engine Failed - status: {pulp.LpStatus[model.status]}')

    eps = 1.e-6  # used to avoid numerical weirdness in the solver's output
    best_price = []
    for t in range(n_periods):
        best_price.append(
            next((prices[p]
                  for p in range(n_prices)
                  if x_var[p, t].varValue > eps and z_var[p, t].varValue > eps), prices[0]))

    return Solution(optimal_prices=best_price, objective=pulp.value(model.objective), status=model.status)
