import numpy as np
import pytest

import revenue_maximization.revenue as pricing_mip


@pytest.mark.parametrize('name,kwargs,expected', [
    (
        'same-revenue-but-different-number-of-booked-rooms',
        {
            'n_available_rooms': 10,
            'demand_matrix': np.repeat([[6, 3, 2]], 3, axis=0).T,
            'prices': [10000, 20000, 30000],
        },
        (
            30000, 180000.0, 1,
        )
    ),
    (
        'small-amount-of-rooms',
        {
            'n_available_rooms': 2,
            'demand_matrix': np.repeat([[6, 3, 2]], 3, axis=0).T,
            'prices': [10000, 20000, 30000],
        },
        (
            30000, 60000.0, 1,
        )
    ),
    (
        'zero-predicted-rooms',
        {
            'n_available_rooms': 10,
            'demand_matrix': np.repeat([[0, 0, 0]], 3, axis=0).T,
            'prices': [10000, 20000, 30000],
        },
        (
            10000, 0.0, 1,
        )
    ),
    (
        'same-revenue-same-number-of-booked-rooms-but-different-frequency',
        {
            'n_available_rooms': 5,
            'demand_matrix': np.repeat([[3, 2, 1]], 3, axis=0).T,
            'prices': [10000, 20000, 30000],
        },
        (
            20000, 110000.0, 1,
        )
    ),
])
def test_pricing_linear_programming_choose_price_results(name, kwargs, expected):
    solution = pricing_mip.optimal_prices(**kwargs)
    assert solution.optimal_prices[0] == expected[0], name
    assert solution.objective == expected[1], name
    assert solution.status == expected[2], name


@pytest.mark.parametrize('name,kwargs,error', [
    (
        'missing_predicted_rooms',
        {
            'n_available_rooms': 5,
            'demand_matrix': np.matrix(()),
            'prices': [],
        },
        {
            'type': ValueError,
            'message': 'needs predictions for calculation',
        }
    ),
    (
        'zero_n_available_rooms',
        {
            'n_available_rooms': 0,
            'demand_matrix': np.repeat([[3, 2, 1]], 3, axis=0).T,
            'prices': [10000, 20000, 30000],
        },
        {
            'type': ValueError,
            'message': 'available rooms should be greater than 0',
        }
    ),
    (
        'negative_n_available_rooms',
        {
            'n_available_rooms': -1,
            'demand_matrix': np.repeat([[3, 2, 1]], 3, axis=0).T,
            'prices': [10000, 20000, 30000],
        },
        {
            'type': ValueError,
            'message': 'available rooms should be greater than 0',
        }
    ),
    (
        'not_enough_prices_in_matrix',
        {
            'n_available_rooms': 1,
            'demand_matrix': np.repeat([[2, 1]], 3, axis=0).T,
            'prices': [10000, 20000, 30000],
        },
        {
            'type': ValueError,
            'message': 'Demand matrix does not match prices length',
        }
    ),
    (
        'non-monotonic-demand-matrix',
        {
            'n_available_rooms': 1,
            'demand_matrix': np.repeat([[3, 2, 3]], 3, axis=0).T,
            'prices': [10000, 20000, 30000],
        },
        {
            'type': ValueError,
            'message': 'Demand matrix has an increase of demand with price.',
        }
    ),
])
def test_pricing_linear_programming_choose_price_error(name, kwargs, error):
    assert error is not None, name

    with pytest.raises(error['type']) as e_info:
        pricing_mip.optimal_prices(**kwargs)
    assert str(e_info.value) == error['message'], name
